@@ -1,5 +1,5 @@
 /*
- * Copyright © 2011-2021 Inria.  All rights reserved.
+ * Copyright © 2011-2023 Inria.  All rights reserved.
  * See COPYING in top-level directory.
  */
 
@@ -17,6 +17,7 @@ int main(void)
   hwloc_obj_t obj, group, saved, res, root;
   hwloc_obj_t objs[32];
   hwloc_uint64_t values[32*32];
+  hwloc_distances_add_handle_t handle;
   int depth;
   hwloc_obj_type_t type;
   unsigned width;
@@ -30,10 +31,23 @@ int main(void)
   hwloc_topology_load(topology);
   root = hwloc_get_root_obj(topology);
   assert(hwloc_topology_get_depth(topology) == 3);
+  /* free instead of inserting */
+  group = hwloc_topology_alloc_group_object(topology);
+  assert(group);
+  group->cpuset = hwloc_bitmap_dup(root->cpuset);
+  err = hwloc_topology_free_group_object(topology, group);
+  assert(err == 0);
+  /* insert without sets, fails */
+  group = hwloc_topology_alloc_group_object(topology);
+  assert(group);
+  res = hwloc_topology_insert_group_object(topology, group);
+  assert(res == NULL);
   /* insert a group identical to root, will be merged */
   group = hwloc_topology_alloc_group_object(topology);
   assert(group);
   group->cpuset = hwloc_bitmap_dup(root->cpuset);
+  err = hwloc_obj_set_subtype(topology, group, "will be merged");
+  assert(!err);
   res = hwloc_topology_insert_group_object(topology, group);
   assert(res);
   assert(res == root);
@@ -44,6 +58,13 @@ int main(void)
   obj = hwloc_get_obj_by_type(topology, HWLOC_OBJ_PACKAGE, 1);
   assert(obj);
   group->cpuset = hwloc_bitmap_dup(obj->cpuset);
+  /* play with its subtype first */
+  err = hwloc_obj_set_subtype(topology, group, NULL);
+  assert(!err);
+  err = hwloc_obj_set_subtype(topology, group, "will be merged");
+  assert(!err);
+  err = hwloc_obj_set_subtype(topology, group, NULL);
+  assert(!err);
   res = hwloc_topology_insert_group_object(topology, group);
   assert(res);
   assert(res == obj);
@@ -57,6 +78,8 @@ int main(void)
   obj = hwloc_get_obj_by_type(topology, HWLOC_OBJ_PU, 12);
   assert(obj);
   hwloc_bitmap_or(group->cpuset, group->cpuset, obj->cpuset);
+  err = hwloc_obj_set_subtype(topology, group, "will fail");
+  assert(!err);
   res = hwloc_topology_insert_group_object(topology, group);
   assert(!res);
   assert(hwloc_topology_get_depth(topology) == 3);
@@ -70,6 +93,8 @@ int main(void)
   assert(obj);
   hwloc_bitmap_or(group->cpuset, group->cpuset, obj->cpuset);
   group->attr->group.kind = (unsigned)-1;
+  err = hwloc_obj_set_subtype(topology, group, "will be merged later");
+  assert(!err);
   res = hwloc_topology_insert_group_object(topology, group);
   assert(res == group);
   saved = group;
@@ -84,9 +109,13 @@ int main(void)
   assert(obj);
   hwloc_bitmap_or(group->cpuset, group->cpuset, obj->cpuset);
   group->attr->group.kind = 0;
+  err = hwloc_obj_set_subtype(topology, group, "will replace previous one");
+  assert(!err);
   res = hwloc_topology_insert_group_object(topology, group);
   assert(res == saved); /* the core should move the contents of this new group into a previous one */
   assert(res != group);
+  assert(res->subtype);
+  assert(!strcmp(res->subtype, "will replace previous one"));
   assert(hwloc_topology_get_depth(topology) == 4);
   /* insert yet another same group with higher kind, it will be dropped in favor of the previous-previous one */
   group = hwloc_topology_alloc_group_object(topology);
@@ -98,9 +127,13 @@ int main(void)
   assert(obj);
   hwloc_bitmap_or(group->cpuset, group->cpuset, obj->cpuset);
   group->attr->group.kind = (unsigned)-1;
+  err = hwloc_obj_set_subtype(topology, group, "will be merged in previous one");
+  assert(!err);
   res = hwloc_topology_insert_group_object(topology, group);
   assert(res == saved);
   assert(res != group);
+  assert(res->subtype);
+  assert(!strcmp(res->subtype, "will replace previous one"));
   assert(hwloc_topology_get_depth(topology) == 4);
   /* insert a conflict group of two packages by nodeset, will fail */
   group = hwloc_topology_alloc_group_object(topology);
@@ -148,9 +181,14 @@ int main(void)
   values[0] = 1; values[1] = 4; values[2] = 4;
   values[3] = 4; values[4] = 1; values[5] = 2;
   values[6] = 4; values[7] = 2; values[8] = 1;
-  err = hwloc_distances_add(topology, 3, objs, values,
-			    HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
-			    HWLOC_DISTANCES_ADD_FLAG_GROUP);
+  handle = hwloc_distances_add_create(topology, NULL,
+                                      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
+                                      0);
+  assert(handle);
+  err = hwloc_distances_add_values(topology, handle, 3, objs, values, 0);
+  assert(!err);
+  err = hwloc_distances_add_commit(topology, handle,
+                                   HWLOC_DISTANCES_ADD_FLAG_GROUP);
   assert(!err);
   /* 1 groups at depth 1 and 2 */
   depth = hwloc_get_type_depth(topology, HWLOC_OBJ_GROUP);
@@ -199,9 +237,14 @@ int main(void)
   values[10] = 4; values[11] = 4; values[12] = 1; values[13] = 4; values[14] = 4;
   values[15] = 4; values[16] = 4; values[17] = 4; values[18] = 1; values[19] = 2;
   values[20] = 4; values[21] = 4; values[22] = 4; values[23] = 2; values[24] = 1;
-  err = hwloc_distances_add(topology, 5, objs, values,
-			    HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
-			    HWLOC_DISTANCES_ADD_FLAG_GROUP);
+  handle = hwloc_distances_add_create(topology, NULL,
+                                      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
+                                      0);
+  assert(handle);
+  err = hwloc_distances_add_values(topology, handle, 5, objs, values, 0);
+  assert(!err);
+  err = hwloc_distances_add_commit(topology, handle,
+                                   HWLOC_DISTANCES_ADD_FLAG_GROUP);
   assert(!err);
   /* 1 node */
   depth = hwloc_get_type_depth(topology, HWLOC_OBJ_NUMANODE);
@@ -274,9 +317,15 @@ int main(void)
   /* group 8cores as 2*2*2 */
   for(i=0; i<16; i++)
     objs[i] = hwloc_get_obj_by_type(topology, HWLOC_OBJ_CORE, i);
-  assert(!hwloc_distances_add(topology, 16, objs, values,
-			      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
-			      HWLOC_DISTANCES_ADD_FLAG_GROUP));
+  handle = hwloc_distances_add_create(topology, NULL,
+                                      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
+                                      0);
+  assert(handle);
+  err = hwloc_distances_add_values(topology, handle, 16, objs, values, 0);
+  assert(!err);
+  err = hwloc_distances_add_commit(topology, handle,
+                                   HWLOC_DISTANCES_ADD_FLAG_GROUP);
+  assert(!err);
   depth = hwloc_topology_get_depth(topology);
   assert(depth == 6);
   width = hwloc_get_nbobjs_by_depth(topology, 0);
@@ -302,9 +351,15 @@ int main(void)
   hwloc_topology_load(topology);
   for(i=0; i<16; i++)
     objs[i] = hwloc_get_obj_by_type(topology, HWLOC_OBJ_CORE, i);
-  assert(!hwloc_distances_add(topology, 16, objs, values,
-			      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
-			      HWLOC_DISTANCES_ADD_FLAG_GROUP|HWLOC_DISTANCES_ADD_FLAG_GROUP_INACCURATE));
+  handle = hwloc_distances_add_create(topology, NULL,
+                                      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
+                                      0);
+  assert(handle);
+  err = hwloc_distances_add_values(topology, handle, 16, objs, values, 0);
+  assert(!err);
+  err = hwloc_distances_add_commit(topology, handle,
+                                   HWLOC_DISTANCES_ADD_FLAG_GROUP|HWLOC_DISTANCES_ADD_FLAG_GROUP_INACCURATE);
+  assert(!err);
   depth = hwloc_topology_get_depth(topology);
   assert(depth == 6);
   hwloc_topology_destroy(topology);
@@ -315,9 +370,15 @@ int main(void)
   hwloc_topology_load(topology);
   for(i=0; i<16; i++)
     objs[i] = hwloc_get_obj_by_type(topology, HWLOC_OBJ_CORE, i);
-  assert(!hwloc_distances_add(topology, 16, objs, values,
-			      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
-			      HWLOC_DISTANCES_ADD_FLAG_GROUP|HWLOC_DISTANCES_ADD_FLAG_GROUP_INACCURATE));
+  handle = hwloc_distances_add_create(topology, NULL,
+                                      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
+                                      0);
+  assert(handle);
+  err = hwloc_distances_add_values(topology, handle, 16, objs, values, 0);
+  assert(!err);
+  err = hwloc_distances_add_commit(topology, handle,
+                                   HWLOC_DISTANCES_ADD_FLAG_GROUP|HWLOC_DISTANCES_ADD_FLAG_GROUP_INACCURATE);
+  assert(!err);
   depth = hwloc_topology_get_depth(topology);
   assert(depth == 6);
   hwloc_topology_destroy(topology);
@@ -327,9 +388,15 @@ int main(void)
   hwloc_topology_load(topology);
   for(i=0; i<16; i++)
     objs[i] = hwloc_get_obj_by_type(topology, HWLOC_OBJ_CORE, i);
-  assert(!hwloc_distances_add(topology, 16, objs, values,
-			      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
-			      HWLOC_DISTANCES_ADD_FLAG_GROUP /* no inaccurate flag, cannot group */));
+  handle = hwloc_distances_add_create(topology, NULL,
+                                      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
+                                      0);
+  assert(handle);
+  err = hwloc_distances_add_values(topology, handle, 16, objs, values, 0);
+  assert(!err);
+  err = hwloc_distances_add_commit(topology, handle,
+                                   HWLOC_DISTANCES_ADD_FLAG_GROUP /* no inaccurate flag, cannot group */);
+  assert(!err);
   depth = hwloc_topology_get_depth(topology);
   assert(depth == 4);
   hwloc_topology_destroy(topology);
@@ -340,9 +407,15 @@ int main(void)
   hwloc_topology_load(topology);
   for(i=0; i<16; i++)
     objs[i] = hwloc_get_obj_by_type(topology, HWLOC_OBJ_CORE, i);
-  assert(!hwloc_distances_add(topology, 16, objs, values,
-			      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
-			      HWLOC_DISTANCES_ADD_FLAG_GROUP|HWLOC_DISTANCES_ADD_FLAG_GROUP_INACCURATE));
+  handle = hwloc_distances_add_create(topology, NULL,
+                                      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
+                                      0);
+  assert(handle);
+  err = hwloc_distances_add_values(topology, handle, 16, objs, values, 0);
+  assert(!err);
+  err = hwloc_distances_add_commit(topology, handle,
+                                   HWLOC_DISTANCES_ADD_FLAG_GROUP|HWLOC_DISTANCES_ADD_FLAG_GROUP_INACCURATE);
+  assert(!err);
   depth = hwloc_topology_get_depth(topology);
   assert(depth == 4);
   hwloc_topology_destroy(topology);
@@ -353,9 +426,15 @@ int main(void)
   hwloc_topology_load(topology);
   for(i=0; i<16; i++)
     objs[i] = hwloc_get_obj_by_type(topology, HWLOC_OBJ_CORE, i);
-  assert(!hwloc_distances_add(topology, 16, objs, values,
-			      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
-			      HWLOC_DISTANCES_ADD_FLAG_GROUP|HWLOC_DISTANCES_ADD_FLAG_GROUP_INACCURATE));
+  handle = hwloc_distances_add_create(topology, NULL,
+                                      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
+                                      0);
+  assert(handle);
+  err = hwloc_distances_add_values(topology, handle, 16, objs, values, 0);
+  assert(!err);
+  err = hwloc_distances_add_commit(topology, handle,
+                                   HWLOC_DISTANCES_ADD_FLAG_GROUP|HWLOC_DISTANCES_ADD_FLAG_GROUP_INACCURATE);
+  assert(!err);
   depth = hwloc_topology_get_depth(topology);
   assert(depth == 4);
   hwloc_topology_destroy(topology);
@@ -387,9 +466,15 @@ int main(void)
       else
         values[i+8*j] = values[j+8*i] = 8;
   }
-  assert(!hwloc_distances_add(topology, 8, objs, values,
-			      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
-			      HWLOC_DISTANCES_ADD_FLAG_GROUP));
+  handle = hwloc_distances_add_create(topology, NULL,
+                                      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
+                                      0);
+  assert(handle);
+  err = hwloc_distances_add_values(topology, handle, 8, objs, values, 0);
+  assert(!err);
+  err = hwloc_distances_add_commit(topology, handle,
+                                   HWLOC_DISTANCES_ADD_FLAG_GROUP);
+  assert(!err);
   depth = hwloc_topology_get_depth(topology);
   assert(depth == 4);
   /* group 4cores as 2*2 */
@@ -403,9 +488,15 @@ int main(void)
       else
         values[i+8*j] = values[j+8*i] = 8;
   }
-  assert(!hwloc_distances_add(topology, 8, objs, values,
-			      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
-			      HWLOC_DISTANCES_ADD_FLAG_GROUP));
+  handle = hwloc_distances_add_create(topology, NULL,
+                                      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
+                                      0);
+  assert(handle);
+  err = hwloc_distances_add_values(topology, handle, 8, objs, values, 0);
+  assert(!err);
+  err = hwloc_distances_add_commit(topology, handle,
+                                   HWLOC_DISTANCES_ADD_FLAG_GROUP);
+  assert(!err);
   depth = hwloc_topology_get_depth(topology);
   assert(depth == 5);
   width = hwloc_get_nbobjs_by_depth(topology, 0);
@@ -431,9 +522,15 @@ int main(void)
       else
         values[i+32*j] = values[j+32*i] = 8;
   }
-  assert(!hwloc_distances_add(topology, 32, objs, values,
-			      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
-			      HWLOC_DISTANCES_ADD_FLAG_GROUP));
+  handle = hwloc_distances_add_create(topology, NULL,
+                                      HWLOC_DISTANCES_KIND_MEANS_LATENCY|HWLOC_DISTANCES_KIND_FROM_USER,
+                                      0);
+  assert(handle);
+  err = hwloc_distances_add_values(topology, handle, 32, objs, values, 0);
+  assert(!err);
+  err = hwloc_distances_add_commit(topology, handle,
+                                   HWLOC_DISTANCES_ADD_FLAG_GROUP);
+  assert(!err);
   depth = hwloc_topology_get_depth(topology);
   assert(depth == 6);
   width = hwloc_get_nbobjs_by_depth(topology, 0);
